/* WordInfo.c -- Copyright 1989 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* WordInfo.c -- handle the database of words for lq-Text.
 * 
 * lq-text keeps a master list of all of the words that have ever been
 * seen.  Currently, this is in dbm format (sdbm or ndbm).
 * For each word, there's an associated WID (a unique number), an offset
 * into the master database (see pblock.c), and possibly thesaurus info.
 *
 * $Id: WordInfo.c,v 2.29 93/11/28 15:13:55 lee Exp $
 *
 */

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <unistd.h>

#include "fileinfo.h"
#include "smalldb.h"
#include "wordindex.h"
#include "wordinfo.h"
#include "numbers.h"

#include "emalloc.h"

#include "wordrules.h" /* max word length */

#include "pblock.h"

/** declarations: **/
/** Unix system calls that need to be declared: **/
extern int open(); /* (these are not the stdio fopen and fclose) */
extern long lseek(); /* watch out for this on 16 bit (286, PDP11) systems! */
extern int read(), write();

/** Unix Library Calls that need to be declared: **/

/** lqtext Library calls that need to be declared: **/
extern void Deletepblock();
extern void lqGetFileModes();
extern t_WordPlace *GetWordPlaces(); /* pblock.c */

/** Functions within this file that need to be declared: **/
t_WordInfo *MakeWordInfo();
void SlayWordInfo();

/** **/

#ifdef ASCIITRACE
extern int AsciiTrace;
#endif

#define WIDINDEXCACHELEN (WIDBLOCKSIZE*32)

#define ui(x) ((unsigned int) x)
#define UL(x) ((unsigned long) x)

static int Widfd = (-1);
static long WidPos = 0L;

PRIVATE void
OpenWordIndexFile()
{
    int Flags, Modes;

    lqGetFileModes(&Flags, &Modes);

    if ((Widfd = open(WidIndexFile, Flags, Modes)) < 0) {
	Error(E_FATAL|E_SYS,
	    "Can't open WID file \"%s\" flags %d mode %d",
	    WidIndexFile, Flags, Modes
	);
    }
    WidPos = 0L;
    (void) GetMaxWID(0);
}

extern t_WID LastNextWIDVal;

PRIVATE unsigned char *
ReadWordInfoIndexBlock(WID)
    t_WID WID;
{
    int i;
    static unsigned char Buffer[WIDBLOCKSIZE + 5];
	/* the +5 allows for overrun in sWriteNumber called from Putlong() */

    if (!WID) {
	Error(E_BUG,
	    "%s: %d: ReadWordInfoIndexBlock(0L) illegal",
	    __FILE__, __LINE__
	);
    }

    if (Widfd < 0) {
	OpenWordIndexFile();
    }

    /* Optimisation: if the WID is greater than the largest allocated WID,
     * there's no point in looking at the file!
     */
    if (LastNextWIDVal && WID > LastNextWIDVal) {
	if (WID > GetMaxWID()) {
	    Error(E_WARN, "WordInfo for %ld, max is %ld, request ignored",
		    WID, LastNextWIDVal
	    );
	    return (unsigned char *) 0;
	}
    }

    if (WidPos != (long) (WID * WIDBLOCKSIZE)) {
	WidPos = (long) (WID * WIDBLOCKSIZE);
	if (lseek(Widfd, WidPos, 0) < 0) {
	    Error(E_FATAL|E_BUG|E_SYS,
		"WID2WordInfo: WID %ld: lseek(%d=\"%s\", %ld, 0) failed",
		WID, Widfd, WidIndexFile, WidPos
	    );
	    return (unsigned char *) 0;
	}
    }

    if ((i = read(Widfd, Buffer, WIDBLOCKSIZE)) != WIDBLOCKSIZE) {
	Error(E_SYS|E_BUG|E_WARN,
	    "Tried to read %d bytes from %d=\"%s\", but got %d",
	    WIDBLOCKSIZE,
	    Widfd,
	    WidIndexFile,
	    i
	);
	WidPos = -1L;
    }

    WidPos += WIDBLOCKSIZE;

    return Buffer;
}

PRIVATE void
WriteWordInfoIndexBlock(WID, Buffer)
    t_WID WID;
    unsigned char *Buffer;
{
    if (!WID) {
	Error(E_BUG,
	    "%s: %d: ReadWordInfoIndexBlock(0L) illegal",
	    __FILE__, __LINE__
	);
    }

    if (Widfd < 0) {
	OpenWordIndexFile();
    }

    if (WidPos != (long) (WID * WIDBLOCKSIZE)) {
	WidPos = (long) (WID * WIDBLOCKSIZE);

	if (lseek(Widfd, WidPos, 0) < 0) {
	    Error(E_SYS|E_FATAL,
		"Index \"%s\": UpdateMatchCount(%lu) lseek (B) to %ld failed",
		WidIndexFile,
		WID,
		WidPos
	    );
	}
    }

    if (write(Widfd, Buffer, WIDBLOCKSIZE) != WIDBLOCKSIZE) {
	Error(E_SYS|E_FATAL,
	    "Index \"%s\": UpdateMatchCount(%lu) write failed",
	    WidIndexFile,
	    WID
	);
    }

    WidPos += WIDBLOCKSIZE;
}

API t_WordInfo *
WID2WordInfo(WID)
    t_WID WID;
{
    unsigned char *q;
    unsigned char *Buffer;
    t_WordInfo *WP;

    Buffer = q = ReadWordInfoIndexBlock(WID);
    if (!q) {
	return (t_WordInfo *) 0;
    }

    if (WordsInWordIndex) {
	unsigned short L;

	if ((L = sReadNumber(&q)) == 0) {
	    Error(E_WARN,
		"WID2WordInfo: Database corrupt, WID %lu has wordlength zero",
		WID
	    );
	    return (t_WordInfo *) 0;
	}
	WP = MakeWordInfo(WID, (int) L, q);
	q += L;
    } else {
	char buf[25];
	(void) sprintf(buf, "W%d", WID);
	WP = MakeWordInfo(WID, (int) strlen(buf), buf);
    }

    WP->Offset = sReadNumber(&q) * BLOCKSIZE;
#ifdef LQ_1_12_COMPAT
    WP->NumberOfWordPlaces = sReadNumber(&q);
#else
    /* WP->NumberOfWordPlaces =
     */
    /* q[0] is  the least significant byte.  What happened to PUT4/GET4? */
    if (WP->Offset != 0L) {
	unsigned long L;

	L = ui(q[3] & 255);
	L <<= 8;
	L |= ui(q[2] & 255);
	L <<= 8;
	L |= ui(q[1] & 255);
	L <<= 8;
	L |= ui(q[0] & 255);

	WP->NumberOfWordPlaces = L;

	q += 4;
    } else {
	WP->NumberOfWordPlaces = sReadNumber(&q);
    }
#endif

    /* Now, maybe read some WordPlace tuplets: */
    Buffer[WIDBLOCKSIZE] = '\0';

    if (q - Buffer < WIDBLOCKSIZE) {
	if (WP->Offset == 0L) {
	    /* In this case, all of the matches, if any,
	     * fit in the index entry, so we might as well deal with
	     * them now.
	     */
	    WP->WordPlaces = GetWordPlaces(
		WP->WID,
		q,
		WIDBLOCKSIZE - (q - Buffer),
		0L,
		WP->NumberOfWordPlaces
	    );
	    WP->DataBlock = (char *) 0;
	    WP->WordPlacesInHere = WP->NumberOfWordPlaces;
	} else {

#if 0
	    /* this will make WID2WordInfo() fetch all of the places;
	     * this is a bad thing, as it results in extra I/O when all
	     * we want is the index entry.  This turns out frequently
	     * to be the case, and in the other cases we may well need
	     * to read the WordPlaces later anyway, e.g. because we're
	     * appending to the index entry and need the last offset,
	     * or becase we want to apply a "Where" function to get fewer
	     * places back, important if there are a large number of hits.
	     */
	    WP->WordPlaces = GetWordPlaces(
		WP->WID,
		q,
		WIDBLOCKSIZE - (q - Buffer),
		WP->Offset,
		WP->NumberOfWordPlaces
	    );
	    WP->WordPlacesInHere = WP->NumberOfWordPlaces;
	    WP->DataBlock = WP->WordPlaceStart = (char *) 0;
#else
	    /* Save the data block so that we can use it later */

	    WP->DataBlock = emalloc(
		"WID data block",
		WIDBLOCKSIZE + 5 /* - (q - Buffer) */
	    );
	    (void) bcopy(
		Buffer,
		WP->DataBlock,
		WIDBLOCKSIZE /* - (q - Buffer) */
	    );
	    /* null terminate the buffer to prevent overun on read: */
	    Buffer[WIDBLOCKSIZE] = '\0';
	    WP->WordPlaceStart = &(WP->DataBlock[q - Buffer]);
	    WP->WordPlaces = (t_WordPlace *) 0;
	    WP->WordPlacesInHere = 0L;
#endif
	}
    } else {
	Error(E_BUG, "block too small for %ld (%s)", WP->WID, WP->Word);
    }

    /* done! */
    return WP;
}

/* Make WordInfo Block Header... */
API void
MkWIBH(WordInfo, pblock)
    t_WordInfo *WordInfo;
    t_pblock *pblock;
{
    char PairBuffer[WIDBLOCKSIZE + 5]; /* the +5 allows for overrun... */
    char *q = PairBuffer;

#ifdef ASCIITRACE
    if (AsciiTrace > 15) {
	fprintf(stderr, "\tMake info block header for %s, Offset %lu==%lu\n",
	WordInfo->Word, pblock->ChainStart, WordInfo->Offset);
    }
#endif

    if (WordsInWordIndex) {
	(void) sWriteNumber(&q, UL(WordInfo->Length), q, WIDBLOCKSIZE);
	(void) strncpy(q, WordInfo->Word, WordInfo->Length);
	q += WordInfo->Length;
    }

    if (pblock && pblock->ChainStart) {
	register unsigned long L;

	sWriteNumber(
	    &q,
	    UL(pblock->ChainStart / BLOCKSIZE),
	    q,
	    WIDBLOCKSIZE - (q - PairBuffer)
	);
#ifdef LQ_1_12_COMPAT
	sWriteNumber(
	    &q,
	    WordInfo->NumberOfWordPlaces,
	    q,
	    WIDBLOCKSIZE - (q - PairBuffer)
	);
#else
	L = WordInfo->NumberOfWordPlaces;

	q[0] = (L & ui(255)); /* least significant */
	L >>= 8;
	q[1] = (L & ui(255));
	L >>= 8;
	q[2] = (L & ui(255));
	L >>= 8;
	q[3] = (L & ui(255)); /* most significant */
	q += 4;
#endif
    } else {
	sWriteNumber(&q, 0L, q, WIDBLOCKSIZE - (q - PairBuffer)); /* offset */
	sWriteNumber(
	    &q,
	    WordInfo->NumberOfWordPlaces,
	    q,
	    WIDBLOCKSIZE - (q - PairBuffer)
	);
    }

    if (WordInfo->DataBlock) {
	efree(WordInfo->DataBlock);
    }
    WordInfo->DataBlock = emalloc(
	"WID Data Block",
	WIDBLOCKSIZE + 5
    );
    (void) bcopy(
	PairBuffer,
	WordInfo->DataBlock,
	WIDBLOCKSIZE 
    );
    WordInfo->WordPlaceStart = &(WordInfo->DataBlock[q - PairBuffer]);
    /* null terminate the buffer to prevent over-run on read: */
    WordInfo->DataBlock[WIDBLOCKSIZE] = '\0';
}

API void
UpdateMatchCount(WID, AddedThese)
    t_WID WID;
    unsigned long AddedThese;
{
    unsigned long Current;
    unsigned long Offset;
    unsigned char *q;
    unsigned char *BlockStart;
    unsigned long Total;

    q = ReadWordInfoIndexBlock(WID);
    if (!q) {
	Error(E_BUG|E_FATAL,
	    "%s: %d: UpdateMatchCount: couldn't read WID block for %lu",
	    __FILE__, __LINE__,
	    WID
	);
    }
    BlockStart = q;

    if (WordsInWordIndex) {
	unsigned short L;

	if ((L = sReadNumber(&q)) == 0) {
	    Error(E_WARN,
		"WID2WordInfo: Database corrupt, WID %lu has wordlength zero",
		WID
	    );
	    return;
	}
	q += L; /* skip over the word */
    }

    /* read the start of the chain */
    Offset = sReadNumber(&q) * BLOCKSIZE;
    if (!Offset) {
	Error(E_BUG|E_FATAL,
	    "%s: %d: UpdateMatchCount(WID %lu, add %lu pairs) but offset is 0",
	    __FILE__, __LINE__,
	    WID, AddedThese
	);
    }

#ifdef LQ_1_12_COMPAT
    Error(E_INTERNAL|E_ABORT|E_BUG|E_FATAL,
	"%s: %d: UpdateMatchCount called in Compatibility mode.",
	__FILE__, __LINE__
    );
#else
    {
	unsigned long L;

	L = ui(q[3] & 255);
	L <<= 8;
	L |= ui(q[2] & 255);
	L <<= 8;
	L |= ui(q[1] & 255);
	L <<= 8;
	L |= ui(q[0] & 255);

	Current = L;
    }
#endif

    Current += AddedThese;
    Total = Current;

    q[0] = (Current & ui(255)); /* least significant */
    Current >>= 8;
    q[1] = (Current & ui(255));
    Current >>= 8;
    q[2] = (Current & ui(255));
    Current >>= 8;
    q[3] = (Current & ui(255));

#ifdef ASCIITRACE
    {
	unsigned long L;

	L = q[3];
	L <<= 8;
	L |= q[2];
	L <<= 8;
	L |= q[1];
	L <<= 8;
	L |= q[0];

	if (L != Total) {
	    Error(E_ABORT|E_BUG|E_FATAL,
		"UpdateMatchCount(%ld, %ld) L %ld != C %ld",
		WID, AddedThese, L, Total
	    );
	}
    }
#endif

    WriteWordInfoIndexBlock(WID, BlockStart);
}

/* Make WordInfo Block ... */
API int
MkWIB(WordInfo, pblock)
    t_WordInfo *WordInfo;
    t_pblock *pblock;
{
    extern unsigned int PutWordPlaces();

    /* See how many pairs from the given pblock fit into WordInfo...
     * and leave them there for later use.
     */

#ifdef ASCIITRACE
    if (AsciiTrace > 3) {
	fprintf(stderr, "MkWIB Make info block for %s/%d at %ld\n",
			    WordInfo->Word, WordInfo->WID, WordInfo->Offset);
    }
#endif

    MkWIBH(WordInfo, pblock);

    if (pblock == (t_pblock *) 0) {
	/* No WordPlaces to put in! */
	WordInfo->WordPlacesInHere = 0;
	return 0;
    }

    if (pblock->ChainStart != 0L) {
	Error(E_WARN,
	    "liblqtext/WordInfo.c::MkWIB() pblock->ChainStart %ld != 0",
	    pblock->ChainStart
	);
    }

    return WordInfo->WordPlacesInHere = PutWordPlaces(
	pblock->WordPlaces,
	WordInfo->WID,
	(unsigned long) 0L, /* 0 means 1st block not in data file */
	WordInfo->DataBlock,
	(unsigned char *) WordInfo->WordPlaceStart,
	WIDBLOCKSIZE,
	0L, /* start */
	0L, /* blocksize */
	pblock->NumberOfWordPlaces
    );
}

API t_WID
Word2WID(Word, Length)
    char *Word;
    unsigned int Length;
{
    DBM *db;
    datum key, data;
    char *q;
    t_WID WID;
 
    if (Length > MaxWordLength) {
	Length = MaxWordLength; /* NOTE: no trailing \0 required. */
    }

    /* contact database server */
    if ((db = startdb(WordIndex)) == (DBM *) 0) {
	Error(E_FATAL|E_SYS,
	    "Word2WID: Couldn't open Word Index (dbm) database \"%s\"",
	    WordIndex
	);
    }

    key.dptr = Word;
    key.dsize = Length;

    data = dbm_fetch(db, key);

    if (data.dptr == (char *) 0 || data.dsize == 0) {
	enddb(db);
	return (t_WID) 0;
    }

    q = data.dptr;
    WID = sReadNumber(&q);
    if (q - data.dptr != data.dsize) {
	Error(E_BUG, "Word2Wid \"%*s\" failed... got %lu", Length, Word, WID);
    }
    if (WID > LastNextWIDVal) {
	(void) GetMaxWID();
    }

    if (WID > LastNextWIDVal) {
	int i;
	int j;

	i *= 3;
	j += 2;
	Error(E_BUG, "Word2WID(%*s) value is %ld, but max WID is %ld! [garbage: %d]",
	    Length,
	    Word,
	    WID,
	    LastNextWIDVal,
	    i + j
	);
    }
    enddb(db);
    return WID;
}
    
API char *
WID2Word(WID)
    t_WID WID;
{
    t_WordInfo *W;
    char *Word;

    if (WID == (t_WID) 0) {
	return (char *) 0;
    }

    if ((W = WID2WordInfo(WID)) == (t_WordInfo *) 0) {
	return (char *) 0;
    }
    Word = W->Word;
    W->Word = (char *) 0;
    SlayWordInfo(W);
    return Word;
}

API t_WID
AllocateWID(Word, Length, WID)
    char *Word;
    int Length;
    t_WID WID;
{
    DBM *db;
    char NumBuf[sizeof(t_WID) * 8/7 + 1];
    char *q = NumBuf;
    datum key, data;

    key.dptr = Word;
    key.dsize = Length;

    sWriteNumber(&q, WID, q, sizeof NumBuf);

    data.dptr = NumBuf;
    data.dsize = q - NumBuf;

    /* contact database server */
    if ((db = startdb(WordIndex)) == (DBM *) 0) {
	Error(E_FATAL|E_SYS,
	    "AllocateWID: Couldn't open dbm Word Index \"%s\"",
	    WordIndex
	);
    }

    if (dbm_store(db, key, data, DBM_REPLACE) < 0) {
	Error(E_FATAL|E_SYS,
	    "WID %ld: dbm_store of %*s failed",
	    WID,
	    Length,
	    Word
	);
    }

    enddb(db);

#ifdef ASCIITRACE
    if (AsciiTrace > 10) {
	t_WID W;

	if ((W = Word2WID(Word, Length)) != WID) {
	    Error(E_BUG,
		"AllocateWID: stored %ld, but retrieved %ld",
		WID,
		W
	    );
	}
    }
#endif

    return WID;
}

API int
PutWordInfoIntoIndex(WordInfo, Offset)
    t_WordInfo *WordInfo;
    unsigned long Offset;
{

    /* Write the physical block to the wordinfo index file.
     * The caller should have allocated a WID with AllocateWID() for
     * this word already, on this or some other program run.
     */

    if (WordInfo->DataBlock == (char *) 0) {
	if (Offset) {
	    Error(E_WARN|E_INTERNAL,
		"%s: %d: WordInfo corrupt for \"%s\"",
		__FILE__, __LINE__,
		WordInfo->Word
	    );
	}
	(void) MkWIB(WordInfo, (t_pblock *) 0);
    }

    WriteWordInfoIndexBlock(WordInfo->WID, WordInfo->DataBlock);

#ifdef ASCIITRACE
    if (AsciiTrace > 10) {
	t_WID w = Word2WID(WordInfo->Word, WordInfo->Length);

	if (w != WordInfo->WID) {
	    Error(E_BUG, "Word %*s Wid changed from %ld to %ld!",
		WordInfo->Length,
		WordInfo->Word,
		WordInfo->WID,
		w
	    );
	}
    }
#endif

    return 0;
}

API int
DeleteWord(Word)
    char *Word;
{
    extern t_pblock *Getpblock();

    t_WID WID;
    t_WordInfo *WordInfo;
    t_pblock *tmp;

    if ((WID = Word2WID(Word, strlen(Word))) == (t_WID) 0) {
	return -1; /* not there */
    }

    /* get info from the list */
    if ((WordInfo = WID2WordInfo(WID)) == (t_WordInfo *) 0) {
	return -1;
    }

    if ((tmp = Getpblock(WordInfo)) != (t_pblock *) NULL) {
	Deletepblock(tmp);
	(void) efree((char *)tmp);
    }

    /* delete the offset from the database, but retain the WID: */
    WordInfo->Offset = 0L;
    WordInfo->NumberOfWordPlaces = 0L;
    WordInfo->WordPlacesInHere = 0;
    PutWordInfoIntoIndex(WordInfo, 0L);
    SlayWordInfo(WordInfo);

    return 0;
}

static t_WordInfo ZeroWordinfo = {
    0,
};

/* Routines to create and destroy WordInfo structures */
API INLINE t_WordInfo *
MakeWordInfo(WID, Length, Word)
    t_WID WID;
    int Length;
    char *Word; /* the word, which might not be nul-terminated */
{
    register t_WordInfo *WP;
    WP = (t_WordInfo *) emalloc("MakeWordInfo", sizeof(t_WordInfo));

    *WP = ZeroWordinfo; /* structure copy */
    WP->WID = WID;

    WP->Word = emalloc("MakeWordInfo.Word", Length + 1);
    (void) strncpy(WP->Word, Word, Length);
    WP->Word[WP->Length = Length] = '\0'; /* strncpy does not add a null */

    return WP;
}

void
API SlayWordInfo(WP)
    t_WordInfo *WP;
{
    if (!WP) return;
    if (WP->Word) efree(WP->Word);
    if (WP->WordPlaces) efree((char *)WP-> WordPlaces);
    if (WP->DataBlock) efree(WP->DataBlock);
    efree((char *) WP);
}

#ifdef ASCIITRACE
API void
fprintWordInfo(stream, W, Caller)
    FILE *stream;
    t_WordInfo *W;
    char *Caller;
{
    fprintf(stream, "%s: WordInfo 0x%x: {\n", Caller, W);
    (void) fflush(stderr);
    if (W) {
	fprintf(stream, "\tWID: %ld (%s, len %u)\n",
			W->WID, W->Word, (unsigned int) W->Length);
	fprintf(stream, "\tNumberOfWordPlaces: %lu In here: %d\n",
			W->NumberOfWordPlaces, W->WordPlacesInHere);
	fprintf(stream, "\tFID: %ld; Offset: %lu\n", W->FID, W->Offset);
	if (W->DataBlock) fprintf(stream, "\tDataBlock: 0x%x\n", W->DataBlock);
	if (W->WordPlaceStart) {
	    fprintf(stream, "\tWordPlaceStart: 0x%x\n", W->WordPlaceStart);
	}
	if (W->WordPlaces) {
	    fprintf(stream, "\tWordPlaces: 0x%x\n", W->WordPlaces);
	}
	if (W->WordPlace.FID && W->WordPlace.FID != W->FID) {
	    fprintf(stream, "\tWordPlace->FID: %ld != FID\n", W->WordPlace.FID);
	}
	fprintf(stream, "\tWordPlace: (Block: %lu; Word %u",
			W->WordPlace.BlockInFile, W->WordPlace.WordInBlock);
	if (W->WordPlace.Flags || W->WordPlace.StuffBefore) {
	    fprintf(stream, "; Flags %u", W->WordPlace.Flags);
	    fprintf(stream, "; StuffBefore: %u", (unsigned int)
					    W->WordPlace.StuffBefore);
	}
	fprintf(stream, ")\n");
	fprintf(stream, "} %s: WordInfo 0x%x\n", Caller, W);
    }
    fflush(stream);
}
#endif /*ASCIITRACE*/

